import logging

from .pi18 import pi18

log = logging.getLogger("pi18sv")

COMMANDS = {
    "PI": {
        "name": "PI",
        "prefix": "^P005",
        "nocrc": True,
        "description": "Device Protocol Version inquiry",
        "help": " -- queries the device protocol version",
        "type": "QUERY",
        "response": [["string", "Protocol Version", ""]],
        "test_responses": [
            b"^D00518;\x03\r",
        ],
    },
    "T": {
        "name": "T",
        "prefix": "^P004",
        "nocrc": True,
        "description": "Query current time",
        "help": " -- queries current time from the Inverter",
        "type": "QUERY",
        "response": [
            ["string", "DateTime", "YYYYMMDDHHMMSS"],
        ],
        "test_responses": [
            b"^D01720210521234743\x0eR\r",
        ],
    },
    "ET": {
        "name": "ET",
        "prefix": "^P005",
        "nocrc": True,
        "description": "Query total generated energy",
        "help": " -- queries total generated energy from the Inverter",
        "type": "QUERY",
        "response": [
            ["int", "Generated energy", "kWh"],
        ],
        "test_responses": [
            b"^D01100006591\xba\x10\r",
        ],
    },
    "EY": {
        "name": "EY",
        "prefix": "^P009",
        "nocrc": True,
        "description": "Query generated energy of year",
        "help": " -- queries generated energy for the year YYYY from the Inverter",
        "type": "QUERYEN",
        "response": [
            ["int", "Generated energy", "Wh"],
        ],
        "test_responses": [
            b"^D01100006591\xba\x10\r",
        ],
        "regex": "EY(\\d\\d\\d\\d)$",
    },
    "EM": {
        "name": "EM",
        "prefix": "^P011",
        "nocrc": True,
        "description": "Query generated energy of month",
        "help": " -- queries generated energy for the month YYYYMM from the Inverter",
        "type": "QUERYEN",
        "response": [
            ["int", "Generated energy", "Wh"],
        ],
        "test_responses": [
            b"^D01000006591\xba\x10\r",
        ],
        "regex": "EM(\\d\\d\\d\\d\\d\\d)$",
    },
    "ED": {
        "name": "ED",
        "prefix": "^P013",
        "description": "Query generated energy of day",
        "help": " -- queries generated energy for the day YYYYMMDD from the Inverter",
        "type": "QUERYEN",
        "response": [
            ["int", "Generated energy", "Wh"],
        ],
        "test_responses": [
            b"^D01100003537\x89X\r",
        ],
        "regex": "ED(\\d\\d\\d\\d\\d\\d\\d\\d)$",
    },
    "ID": {
        "name": "ID",
        "prefix": "^P005",
        "nocrc": True,
        "description": "Device Serial Number inquiry",
        "help": " -- queries the device serial number",
        "type": "QUERY",
        "response": [["string", "Serial Number", ""]],
        "test_responses": [
            b"^D0251496161704100242000000le\r",
        ],
    },
    "VFW": {
        "name": "VFW",
        "prefix": "^P006",
        "description": "Device CPU version inquiry",
        "help": " -- queries the CPU version",
        "type": "QUERY",
        "response": [
            ["int", "Main CPU Version", ""],
            ["int", "Slave 1 CPU version", ""],
            ["int", "Slave 2 CPU version", ""],
        ],
        "test_responses": [
            b"^D02005402,08025,00000)\xc3\x9d\r",
        ],
    },
    "MCHGCR": {  # On my SV IV, this commands returns 12 values, from 10 to 120 amps
        "name": "MCHGCR",
        "prefix": "^P009",
        "description": "Query Max. charging current selectable values",
        "help": " -- queries the Max. charging current selectable values",
        "type": "QUERY",
        "response": [
            ["int", "Max. charging current selectable value 1", "A"],
            ["int", "Max. charging current selectable value 2", "A"],
            ["int", "Max. charging current selectable value 3", "A"],
            ["int", "Max. charging current selectable value 4", "A"],
            ["int", "Max. charging current selectable value 5", "A"],
            ["int", "Max. charging current selectable value 6", "A"],
            ["int", "Max. charging current selectable value 7", "A"],
        ],
        #                     ["int", "Max. charging current selectable value 8", "A"]],
        #                     ["int", "Max. charging current selectable value 9", "A"]],
        #                     ["int", "Max. charging current selectable value 10", "A"]],
        #                     ["int", "Max. charging current selectable value 11", "A"]],
        #                     ["int", "Max. charging current selectable value 12", "A"]],
        "test_responses": [
            b"^D030010,020,030,040,050,060,070\n",
        ],
    },
    "MUCHGCR": {  # On my SV IV, this commands returns 13 values, from 2 to 120 amps
        "name": "MUCHGCR",
        "prefix": "^P010",
        "description": "Query Max. AC charging current selectable values",
        "help": " -- queries the Max. AC charging current selectable values",
        "type": "QUERY",
        "response": [
            ["int", "Max. charging current selectable value 1", "A"],
            ["int", "Max. charging current selectable value 2", "A"],
            ["int", "Max. charging current selectable value 3", "A"],
            ["int", "Max. charging current selectable value 4", "A"],
            ["int", "Max. charging current selectable value 5", "A"],
            ["int", "Max. charging current selectable value 6", "A"],
            ["int", "Max. charging current selectable value 7", "A"],
        ],
        #                     ["int", "Max. charging current selectable value 8", "A"]],
        #                     ["int", "Max. charging current selectable value 9", "A"]],
        #                     ["int", "Max. charging current selectable value 10", "A"]],
        #                     ["int", "Max. charging current selectable value 11", "A"]],
        #                     ["int", "Max. charging current selectable value 12", "A"]],
        #                     ["int", "Max. charging current selectable value 12", "A"]],
        "test_responses": [
            b"^D030010,020,030,040,050,060,070\n",
        ],
    },
    "PRI": {
        "name": "PRI",
        "prefix": "^P007",
        "description": "Query different rated information of parallel system",
        "help": " -- queries different rated information of parallel system",
        "type": "QUERYEN",
        "response": [
            ["option", "Parallel ID connection status", ["Not existent", "Existent"]],
            ["int", "Serial Number valid length", ""],
            ["string", "Serial Number", ""],
            [
                "option",
                "Charging source priority",
                ["Solar first", "Solar and Utility", "Only Solar"],
            ],
            ["int", "Max. charging current", "A"],
            ["int", "Max. AC charging current", "A"],
            [
                "option",
                "Output mode setting",
                [
                    "Single mode",
                    "parallel output",
                    "Phase 1 of 3 phases",
                    "Phase 2 of 3 phases",
                    "Phase 3 of 3 phases",
                ],
            ],
        ],
        "test_responses": [
            b"^D0401,14,96132206100410000000,2,060,050,1Jl\r",
        ],
        "regex": "PRI(\\d)$",
    },
    "PGS": {
        "name": "PGS",
        "prefix": "^P007",
        "description": "Query general status of parallel system",
        "help": " -- queries  general status of parallel system",
        "type": "QUERY",
        "response": [
            ["option", "Parallel ID connection status", ["Not existent", "Existent"]],
            ["int", "Work mode", ""],
            ["int", "Fault code", ""],
            ["int", "Grid voltage", "0.1V"],
            ["int", "Grid frequency", "0.1HZ"],
            ["int", "AC output voltage", "0.1V"],
            ["int", "AC output frequency", "0.1Hz"],
            ["int", "AC output apparent power", "VA"],
            ["int", "AC output active power", "W"],
            ["int", "Total AC output apparent power", "VA"],
            ["int", "Total AC output active power", "W"],
            ["int", "Output load percent", "%"],
            ["int", "Total output load percent", "%"],
            ["int", "Battery voltage", "0.1V"],
            ["int", "Battery discharge current", "A"],
            ["int", "Battery charging current", "A"],
            ["int", "Total battery charging current", "A"],
            ["int", "Battery capacity", "%"],
            ["int", "PV1 Input power", "W"],
            ["int", "PV2 Input power", "W"],
            ["int", "PV1 Input voltage", "0.1V"],
            ["int", "PV2 Input voltage", "0.1V"],
            [
                "option",
                "MPPT1 charger status",
                ["abnormal", "normal but not charging", "charging"],
            ],
            [
                "option",
                "MPPT2 charger status",
                ["abnormal", "normal but not charging", "charging"],
            ],
            ["option", "Load connection", ["disconnect", "connect"]],
            ["option", "Battery power direction", ["donothing", "charge", "discharge"]],
            ["option", "DC/AC power direction", ["donothing", "AC-DC", "DC-AC"]],
            ["option", "Line power direction", ["donothing", "input", "output"]],
            ["int", "Max Temperature", "°C"],
        ],
        "test_responses": [
            b"^D1131,5,00,2387,499,2387,499,0716,0165,01232,00505,012,004,518,000,015,031,036,1137,0000,2088,0000,2,0,1,1,2,1,037\x8a$\r",
        ],
        "regex": "PGS(\\d)$",
    },
    "FWS": {
        "name": "FWS",
        "prefix": "^P005",
        "description": "Query fault and warning status",
        "help": " -- queries the fault and warning status",
        "type": "QUERY",
        "response": [
            ["int", "Fault code", ""],
            ["int", "Line fail", ""],
            ["int", "Output circuit short", ""],
            ["int", "Inverter over temperature", ""],
            ["int", "Fan lock", ""],
            ["int", "Battery voltage high", ""],
            ["int", "Battery low", ""],
            ["int", "Battery under", ""],
            ["int", "Over load", ""],
            ["int", "EEProm fail", ""],
            ["int", "Power limit", ""],
            ["int", "PV1 voltage high", ""],
            ["int", "PV2 voltage high", ""],
            ["int", "MPPT1 overload warning", ""],
            ["int", "MPPT2 overload warning", ""],
            ["int", "Battery too low to charge for SCC1", ""],
            ["int", "Battery too low to charge for SCC2", ""],
        ],
        "test_responses": [
            b"^D03900,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0\xc2\xb6\x9e\r",
        ],
    },
    "DI": {
        "name": "DI",
        "prefix": "^P005",
        "nocrc": True,
        "description": "Query default value of changeable parameters",
        "help": "",
        "type": "QUERY",
        "response": [
            ["int", "AC output voltage", "0.1V"],
            ["int", "AC output frequency", "0.1Hz"],
            ["option", "AC input voltage range", ["Appliance", "UPS"]],
            ["int", "Battery Under voltage", "0.1V"],
            ["int", "Charging float voltage", "0.1V"],
            ["int", "Charging bulk voltage", "0.1V"],
            ["int", "Battery default re-charge voltage", "0.1V"],
            ["int", "Battery re-discharge voltage", "0.1V"],
            ["int", "Max charging current", "A"],
            ["int", "Max AC charging current", "A"],
            ["option", "Battery type", ["AGM", "Flooded", "User"]],
            [
                "option",
                "Output source priority",
                ["Solar-Utility-Battery", "Solar-Battery-Utility"],
            ],
            [
                "option",
                "Charger source priority",
                ["Solar first", "Solar and Utility", "Only solar"],
            ],
            [
                "option",
                "Solar power priority",
                ["Battery-Load-Utility", "Load-Battery-Utility"],
            ],
            ["option", "Machine type", ["Off-Grid", "Grid-Tie"]],
            [
                "option",
                "Output mode setting",
                [
                    "Single mode",
                    "parallel output",
                    "Phase 1 of 3 phases",
                    "Phase 2 of 3 phases",
                    "Phase 3 of 3 phases",
                ],
            ],
            ["option", "Silence buzzer or open buzzer", ["disable", "enable"]],
            ["option", "Overload restart", ["disable", "enable"]],
            ["option", "Over temperature restart", ["disable", "enable"]],
            ["option", "LCD backlight on", ["disable", "enable"]],
            ["option", "Alarm when primary source interrupted", ["disable", "enable"]],
            ["option", "Fault code record", ["disable", "enable"]],
            ["option", "Overload bypass", ["disable", "enable"]],
            [
                "option",
                "LCD display escape to default page after timeout",
                ["disable", "enable"],
            ],
        ],
        "test_responses": [
            "^D0682300,500,0,408,540,564,460,540,060,30,0,0,1,0,0,0,1,0,0,1,1,0,1,1Q\xc3\xa6\r"
        ],
    },
    "PIRI": {
        "name": "PIRI",
        "prefix": "^P007",
        "description": "Device rated information",
        "help": " -- queries rated information",
        "type": "QUERY",
        "response": [
            ["int", "AC input rated voltage", "0.1V"],
            ["int", "AC input rated current", "0.1A"],
            ["int", "AC output rated voltage", "0.1V"],
            ["int", "AC output rated frequency", "0.1HZ"],
            ["int", "AC output rated current", "0.1A"],
            ["int", "AC output rated apparent power", "1VA"],
            ["int", "AC output rated active power", "W"],
            ["int", "Battery rated voltage", "0.1V"],
            ["int", "Battery recharge voltage", "0.1V"],
            ["int", "Battery redischarge voltage", "0.1V"],
            ["int", "Battery under voltage", "0.1V"],
            ["int", "Battery bulk voltage", "0.1V"],
            ["int", "Battery float voltage", "0.1V"],
            ["option", "Battery Type", ["AGM", "Flooded", "User"]],
            ["int", "Max AC charging current", "A"],
            ["int", "Max charging current", "A"],
            ["option", "Input voltage range", ["Appliance", "UPS"]],
            [
                "option",
                "Output source priority",
                ["Solar-Utility-Battery", "Solar and Utility"],
            ],
            [
                "option",
                "Charger source priority",
                ["Solar first", "Solar and Utility", "Only Solar"],
            ],
            ["int", "Parallel max num", ""],
            ["option", "Machine type", ["Off-grid", "Grid-tie"]],
            ["option", "Topology", ["transformerless", "transformer"]],
            [
                "option",
                "Output model setting",
                [
                    "Single module",
                    "Parallel output",
                    "Phase 1 of 3 output",
                    "Phase 2 of 3 output",
                    "Phase 3 of 3 output",
                ],
            ],
            [
                "option",
                "Solar power priority",
                ["Battery-Load-Utility", "Load-Battery-Utility"],
            ],
            ["int", "MPPT string", ""],
        ],
        "test_responses": [
            b"^D0892300,243,2300,500,243,5600,5600,480,500,530,480,570,544,2,050,060,0,1,2,9,0,0,1,1,1,00\x8e~\r"
        ],
    },
    "GS": {
        "name": "GS",
        "prefix": "^P005",
        "nocrc": True,  # For some strange reason, crc makes this query return only partial infos ?!
        "description": "Query general status",
        "help": " -- queries general status",
        "type": "QUERY",
        "response": [
            ["int", "AC Input Voltage", "0.1V"],
            ["int", "AC Input Frequency", "0.1Hz"],
            ["int", "AC Output Voltage", "0.1V"],
            ["int", "AC Output Frequency", "0.1Hz"],
            ["int", "AC Output Apparent Power", "VA"],
            ["int", "AC Output Active Power", "W"],
            ["int", "AC Output Load", "%"],
            ["int", "Battery Voltage", "0.1V"],
            ["int", "Battery Voltage from SCC", "0.1V"],
            ["int", "Battery Voltage from SCC2", "0.1V"],
            ["int", "Battery Discharge Current", "A"],
            ["int", "Battery Charge Current", "A"],
            ["int", "Battery Capacity", "%"],
            ["int", "Inverter Temperature", "°C"],
            ["int", "MPPT1 Charger Temperature", "°C"],
            ["int", "MPPT2 Charger Temperature", "°C"],
            ["int", "PV1 Input Power", "W"],
            ["int", "PV2 Input Power", "W"],
            ["int", "PV1 Input Voltage", "0.1V"],
            ["int", "PV2 Input Voltage", "0.1V"],
            [
                "option",
                "Setting value configuration state",
                ["Nothing changed", "Something changed"],
            ],
            [
                "option",
                "MPPT1 charger status",
                ["abnormal", "normal but not charged", "normal"],
            ],
            [
                "option",
                "MPPT2 charger status",
                ["abnormal", "normal but not charged", "normal"],
            ],
            ["option", "Load connection", ["disconnect", "connect"]],
            ["option", "Battery power direction", ["donothing", "charge", "discharge"]],
            ["option", "DC/AC power direction", ["donothing", "AC-DC", "DC-AC"]],
            ["option", "Line power direction", ["donothing", "input", "output"]],
            ["int", "Local parallel ID", ""],
        ],
        "test_responses": [
            b"^D1062336,499,2336,499,0443,0255,007,524,000,000,000,017,042,039,000,000,1326,0000,1907,0000,0,2,0,1,1,2,1,1`\xc3\xaa\r"
        ],
    },
    "MOD": {
        "name": "MOD",
        "prefix": "^P006",
        "description": "Device working mode inquiry",
        "help": " -- queries the device working mode",
        "type": "QUERY",
        "response": [
            [
                "keyed",
                "Working mode",
                {
                    "00": "Power on mode",
                    "01": "Standby mode",
                    "02": "Bypass mode",
                    "03": "Battery mode",
                    "04": "Fault mode",
                    "05": "Hybrid mode (Line mode, Grid mode)",
                },
            ],
        ],
        "test_responses": [
            b"^D00505\xc3\x99\x9f\r",
        ],
    },
    "FLAG": {
        "name": "FLAG",
        "prefix": "^P007",
        "description": "Query enable/disable flag status",
        "help": " -- queries enable/disable flag status from the Inverter",
        "type": "QUERY",
        "response": [
            ["option", "Mute buzzer beep", ["Disabled", "Enabled"]],
            ["option", "Overload bypass function", ["Disabled", "Enabled"]],
            [
                "option",
                "LCD Display escape to default page after timeout",
                ["Disabled", "Enabled"],
            ],
            ["option", "Overload restart", ["Disabled", "Enabled"]],
            ["option", "Over temperature restart", ["Disabled", "Enabled"]],
            ["option", "Backlight on", ["Disabled", "Enabled"]],
            [
                "option",
                "Alarm on when primary source interrupted",
                ["Disabled", "Enabled"],
            ],
            ["option", "Fault code record", ["Disabled", "Enabled"]],
            ["option", "Reserved", ["Disabled", "Enabled"]],
        ],
        "test_responses": [
            b"'^D0201,0,1,0,0,1,1,0,0\xc3\x85\xa0\r'",
        ],
    },
    "ACCT": {
        "name": "ACCT",
        "prefix": "^P005",
        "description": "Query AC charge time bucket",
        "help": " -- queries AC charge time bucket",
        "type": "QUERY",
        "response": [
            ["string", "Start time for enable AC charger working", "HH:MM"],
            ["string", "Ending time for enable AC charger working", "HH:MM"],
        ],
        "test_responses": [
            b"^D0120000,0000\xc3\x82\x1c\r",
        ],
    },
    "ACLT": {
        "name": "ACLT",
        "prefix": "^P005",
        "description": "Query AC supply load time bucket",
        "help": " -- queries AC supply load time bucket",
        "type": "QUERY",
        "response": [
            ["string", "Start time for enable AC supply the load", "HH:MM"],
            ["string", "Ending time for enable AC supply the load", "HH:MM"],
        ],
        "test_responses": [
            b"^D0120000,0000\xc3\x82\x1c\r",
        ],
    },
}
SETTER_COMMANDS = {
    # Now the setters
    "LON": {
        "name": "LON",
        "prefix": "^S007",
        "description": "Set enable/disable machine supply power to the loads",
        "help": " -- examples: LON1 (0: disable, 1: enable)",
        "type": "SETTER",
        "response": [
            ["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}],
        ],
        "test_responses": [
            b"^1\x0b\xc2\r",
            b"^0\x1b\xe3\r",
        ],
        "regex": "LON([01])$",
    },
    "PEA": {
        "name": "PEA",
        "prefix": "^S006",
        "description": "Enable Silence buzzer or open buzzer",
        "help": " -- Enable Silence buzzer or open buzzer",
        "type": "SETTER",
        "response": [
            ["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}],
        ],
        "test_responses": [
            b"^1\x0b\xc2\r",
            b"^0\x1b\xe3\r",
        ],
    },
    "PDA": {
        "name": "PDA",
        "prefix": "^S006",
        "description": "Disable Silence buzzer or open buzzer",
        "help": " -- Disable Silence buzzer or open buzzer",
        "type": "SETTER",
        "response": [
            ["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}],
        ],
        "test_responses": [
            b"^1\x0b\xc2\r",
            b"^0\x1b\xe3\r",
        ],
    },
    "PEB": {
        "name": "PEB",
        "prefix": "^S006",
        "description": "Enable Overload bypass function",
        "help": " -- Enable Overload bypass function",
        "type": "SETTER",
        "response": [
            ["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}],
        ],
        "test_responses": [
            b"^1\x0b\xc2\r",
            b"^0\x1b\xe3\r",
        ],
    },
    "PDB": {
        "name": "PDB",
        "prefix": "^S006",
        "description": "Disable Overload bypass function",
        "help": " -- Disable Overload bypass function",
        "type": "SETTER",
        "response": [
            ["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}],
        ],
        "test_responses": [
            b"^1\x0b\xc2\r",
            b"^0\x1b\xe3\r",
        ],
    },
    "PEC": {
        "name": "PEC",
        "prefix": "^S006",
        "description": "Enable LCD display escape to default page after 1min timeout",
        "help": " -- Enable LCD display escape to default page after 1min timeout",
        "type": "SETTER",
        "response": [
            ["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}],
        ],
        "test_responses": [
            b"^1\x0b\xc2\r",
            b"^0\x1b\xe3\r",
        ],
    },
    "PDC": {
        "name": "PDC",
        "prefix": "^S006",
        "description": "Disable LCD display escape to default page after 1min timeout",
        "help": " -- Disable LCD display escape to default page after 1min timeout",
        "type": "SETTER",
        "response": [
            ["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}],
        ],
        "test_responses": [
            b"^1\x0b\xc2\r",
            b"^0\x1b\xe3\r",
        ],
    },
    "PED": {
        "name": "PED",
        "prefix": "^S006",
        "description": "Enable Overload restart",
        "help": " -- Enable Overload restart",
        "type": "SETTER",
        "response": [
            ["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}],
        ],
        "test_responses": [
            b"^1\x0b\xc2\r",
            b"^0\x1b\xe3\r",
        ],
    },
    "PDD": {
        "name": "PDD",
        "prefix": "^S006",
        "description": "Disable Overload restart",
        "help": " -- Disable Overload restart",
        "type": "SETTER",
        "response": [
            ["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}],
        ],
        "test_responses": [
            b"^1\x0b\xc2\r",
            b"^0\x1b\xe3\r",
        ],
    },
    "PEE": {
        "name": "PEE",
        "prefix": "^S006",
        "description": "Enable Over temperature restart",
        "help": " -- Enable Over temperature restart",
        "type": "SETTER",
        "response": [
            ["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}],
        ],
        "test_responses": [
            b"^1\x0b\xc2\r",
            b"^0\x1b\xe3\r",
        ],
    },
    "PDE": {
        "name": "PDE",
        "prefix": "^S006",
        "description": "Disable Over temperature restart",
        "help": " -- Disable Over temperature restart",
        "type": "SETTER",
        "response": [
            ["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}],
        ],
        "test_responses": [
            b"^1\x0b\xc2\r",
            b"^0\x1b\xe3\r",
        ],
    },
    "PEF": {
        "name": "PEF",
        "prefix": "^S006",
        "description": "Enable Backlight on",
        "help": " -- Enable Backlight on",
        "type": "SETTER",
        "response": [
            ["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}],
        ],
        "test_responses": [
            b"^1\x0b\xc2\r",
            b"^0\x1b\xe3\r",
        ],
    },
    "PDF": {
        "name": "PDF",
        "prefix": "^S006",
        "description": "Disable Backlight on",
        "help": " -- Disable Backlight on",
        "type": "SETTER",
        "response": [
            ["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}],
        ],
        "test_responses": [
            b"^1\x0b\xc2\r",
            b"^0\x1b\xe3\r",
        ],
    },
    "PEG": {
        "name": "PEG",
        "prefix": "^S006",
        "description": "Enable Alarm on when primary source interrupt",
        "help": " -- Enable Alarm on when primary source interrupt",
        "type": "SETTER",
        "response": [
            ["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}],
        ],
        "test_responses": [
            b"^1\x0b\xc2\r",
            b"^0\x1b\xe3\r",
        ],
    },
    "PDG": {
        "name": "PDG",
        "prefix": "^S006",
        "description": "Disable Alarm on when primary source interrupt",
        "help": " -- Disable Alarm on when primary source interrupt",
        "type": "SETTER",
        "response": [
            ["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}],
        ],
        "test_responses": [
            b"^1\x0b\xc2\r",
            b"^0\x1b\xe3\r",
        ],
    },
    "PEH": {
        "name": "PEH",
        "prefix": "^S006",
        "description": "Enable Fault code record",
        "help": " -- Enable Fault code record",
        "type": "SETTER",
        "response": [
            ["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}],
        ],
        "test_responses": [
            b"^1\x0b\xc2\r",
            b"^0\x1b\xe3\r",
        ],
    },
    "PDH": {
        "name": "PDH",
        "prefix": "^S006",
        "description": "Disable Fault code record",
        "help": " -- Disable Fault code record",
        "type": "SETTER",
        "response": [
            ["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}],
        ],
        "test_responses": [
            b"^1\x0b\xc2\r",
            b"^0\x1b\xe3\r",
        ],
    },
    "PEI": {
        "name": "PEI",
        "prefix": "^S006",
        "description": "Set Machine type Grid-Tie",
        "help": " --Set Machine type Grid-Tie",
        "type": "SETTER",
        "response": [
            ["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}],
        ],
        "test_responses": [
            b"^1\x0b\xc2\r",
            b"^0\x1b\xe3\r",
        ],
    },
    "PDI": {
        "name": "PDI",
        "prefix": "^S006",
        "description": "Set Machine type Off-Grid",
        "help": " -- Set Machine type Off-Grid",
        "type": "SETTER",
        "response": [
            ["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}],
        ],
        "test_responses": [
            b"^1\x0b\xc2\r",
            b"^0\x1b\xe3\r",
        ],
    },
    "PF": {
        "name": "PF",
        "prefix": "^S005",
        "description": "Set changeable parameter restore to default value",
        "help": " -- Set changeable parameter restore to default value",
        "type": "SETTER",
        "response": [
            ["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}],
        ],
        "test_responses": [
            b"^1\x0b\xc2\r",
            b"^0\x1b\xe3\r",
        ],
    },
    "MCHGC": {
        "name": "MCHGC",
        "prefix": "^S013",
        "nocrc": True,
        "description": "Set battery maximum charge current",
        "help": " -- MCHGCRm,nnn with m : 0~Parallel number, if single model, it should be 0 ; n : current. Must choose a seable value returned by MCHGCR ; Ex: MCHGC0,050",
        "type": "SETTER",
        "response": [
            ["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}],
        ],
        "test_responses": [
            b"^1\x0b\xc2\r",
            b"^0\x1b\xe3\r",
        ],
        "regex": "MCHGC(\\d,\\d\\d\\d)$",
    },
    "MUCHGC": {  # Not working
        "name": "MUCHGC",
        "prefix": "^S014",
        "nocrc": True,
        "description": "Set battery maximum AC charge current",
        "help": " -- MUCHGCm,nnn with m : 0~Parallel number, if single model, it should be 0 ; n : current. Must choose a seable value returned by MUCHGCR ; Ex: MUCHGC0,050",
        "type": "SETTER",
        "response": [
            ["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}],
        ],
        "test_responses": [
            b"^1\x0b\xc2\r",
            b"^0\x1b\xe3\r",
        ],
        "regex": "MUCHGC(\\d,\\d\\d\\d)$",
    },
    "MCHGV": {
        "name": "MCHGV",
        "prefix": "^S015",
        "description": "Set battery maximum charge voltage",
        "help": " -- MCHGVmmm,nnn with mmm battery constant charge voltage and nnn battery float voltage. Unit 0.1V",
        "type": "SETTER",
        "response": [
            ["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}],
        ],
        "test_responses": [
            b"^1\x0b\xc2\r",
            b"^0\x1b\xe3\r",
        ],
        "regex": "MCHGV(\\d\\d\\d,\\d\\d\\d)$",
    },
    "DAT": {
        "name": "DAT",
        "prefix": "^S018",
        "description": "Set date time",
        "help": " -- examples: DAT190518224530(YYMMDDHHMMSS-12digits)",
        "type": "SETTER",
        "response": [
            ["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}],
        ],
        "test_responses": [
            b"^1\x0b\xc2\r",
            b"^0\x1b\xe3\r",
        ],
        "regex": "DAT(\\d\\d\\d\\d\\d\\d\\d\\d\\d\\d\\d\\d)$",
    },
    "POP": {
        "name": "POP",
        "prefix": "^S007",
        "description": "Set output source priority",
        "help": " -- POP0 : Solar-Utility-Battery ; POP1 : Solar-Battery-Utility",
        "type": "SETTER",
        "response": [
            ["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}],
        ],
        "test_responses": [
            b"^1\x0b\xc2\r",
            b"^0\x1b\xe3\r",
        ],
        "regex": "POP(\\d)$",
    },
    "BUCD": {
        "name": "BUCD",
        "prefix": "^S014",
        "description": "Battery re-charged and re-discharged voltage when utility is available",
        "help": " -- BUCDmmm,nnn with mmm battery recharge voltage when utility available (44 to 51V) and nnn Battery re-discharged voltage when utility is available (0,48-58)",
        "type": "SETTER",
        "response": [
            ["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}],
        ],
        "test_responses": [
            b"^1\x0b\xc2\r",
            b"^0\x1b\xe3\r",
        ],
        "regex": "BUCD(\\d\\d\\d,\\d\\d\\d)$",
    },
}


class pi18sv(pi18):
    def __str__(self):
        return "PI18 protocol handler for Easun SV inverters"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__()
        self._protocol_id = b"PI18SV"
        self.COMMANDS = COMMANDS
        self.COMMANDS.update(SETTER_COMMANDS)
        self.STATUS_COMMANDS = [
            "PI",
            "T",
            "ET",
            "EY",
            "EM",
            "ED",
            "ID",
            "VFW",
            "PIRI",
            "GS",
            "MOD",
            "FWS",
            "FLAG",
            "DI",
            "CHGCR",
            "MUCHGCR",
            "PRI",
            "PGS",
            "ACCT",
            "ACLT",
        ]
        self.SETTINGS_COMMANDS = [
            "LON",
            "PEA",
            "PDA",
            "PEB",
            "PDB",
            "PEC",
            "PDC",
            "PED",
            "PDD",
            "PEE",
            "PDE",
            "PEF",
            "PDF",
            "PEG",
            "PDG",
            "PEH",
            "PDH",
            "PF",
            "MUCHGC",
            "MUCHGV",
            "MCHGV",
            "BUCD",
            "DAT",
        ]
        self.DEFAULT_COMMAND = "GS"


# Bugs / Not working
# EMYYYYDD returns an error



# !/usr/bin/python3
import logging
from argparse import ArgumentParser
from collections import deque
from time import sleep, time
from datetime import date, timedelta  # noqa: F401

import yaml

from mppsolar.libs.mqttbrokerc import MqttBroker
from mppsolar.outputs import output_results
from mppsolar.ports import get_port
from mppsolar.protocols import get_protocol
from mppsolar.version import __version__  # noqa: F401

# from mppsolar.inout import get_port


# Set-up logger
log = logging.getLogger("")


class ConfigError(Exception):
    pass


sample_config = """
  port:
    path: /dev/ttyUSB0
    type: test
    baud: 2400
  protocol: PI30
  commands:
    - command: QPI
      outputs:
      - name: screen
  loop: once
"""

ADHOC_COMMANDS = deque([])
# SPLIT_TOKEN = ","


def mqtt_callback(client, userdata, msg):
    log.info(f"Received `{msg.payload}` on topic `{msg.topic}`")
    # TODO: define message format and extract command and config
    newCommand = msg.payload
    ADHOC_COMMANDS.append(newCommand)


def readConfigFile(configFile=None):
    _config = {}
    if configFile is not None:
        try:
            with open(configFile, "r") as stream:
                _config = yaml.safe_load(stream)
        except yaml.YAMLError as exc:
            log.error(f"Error processing config file: {exc}")
        except FileNotFoundError as exc:
            log.error(f"Error opening config file: {exc}")
    return _config


def processCommandLineOverrides(args):
    _config = {}
    if args.once:
        _config["loop"] = "once"
    if args.info:
        _config["debuglevel"] = logging.INFO
    if args.debug:
        _config["debuglevel"] = logging.DEBUG
    return _config


def main():
    description = f"Power Device Monitoring Utility, version: {__version__}"
    parser = ArgumentParser(description=description)

    parser.add_argument(
        "-C",
        "--configFile",
        nargs="?",
        type=str,
        help="Full location of config file",
        const="./powermon.yaml",
        default=None,
    )
    parser.add_argument(
        "-v", "--version", action="store_true", help="Display the version"
    )
    parser.add_argument(
        "-d",
        "--dumpConfig",
        action="store_true",
        help="Print the config based on options supplied (including the defaults, config file and any overrides)",
    )
    parser.add_argument(
        "-1",
        "--once",
        action="store_true",
        help="Only loop through config once",
    )
    parser.add_argument(
        "-D",
        "--debug",
        action="store_true",
        help="Enable Debug and above (i.e. all) messages",
    )
    parser.add_argument(
        "-I", "--info", action="store_true", help="Enable Info and above level messages"
    )
    parser.add_argument(
        "--adhoc",
        type=str,
        default=None,
        help="Send adhoc command to mqtt adhoc command queue - needs config file specified and populated",
    )

    args = parser.parse_args()
    # prog_name = parser.prog

    # Temporarily set debug level based on command line options
    log.setLevel(logging.WARNING)
    if args.info:
        log.setLevel(logging.INFO)
    if args.debug:
        log.setLevel(logging.DEBUG)

    # Display version if asked
    log.info(description)
    if args.version:
        print(description)
        return None

    # Build configuration from defaults, config file and command line overrides
    log.info(f"Using config file: {args.configFile}")
    # build config - start with defaults
    config = yaml.safe_load(sample_config)
    # build config - update with details from config file
    config.update(readConfigFile(args.configFile))
    # build config - override with any command line arguments
    config.update(processCommandLineOverrides(args))

    # logging (DEBUG, INFO, WARNING, ERROR, CRITICAL)
    log.setLevel(config.get("debuglevel", logging.WARNING))

    # if generateConfigFile is true then print config out
    if args.dumpConfig:
        print("# yaml config for powermon")
        print("# default location /etc/mpp-solar/powermon.yml")
        print(yaml.dump(config))
        return

    # debug dump config
    log.info(f"config: {config}")

    # Build mqtt broker
    mqttconfig = config.get("mqttbroker", {})
    mqtt_broker = MqttBroker(config=mqttconfig)
    # is this just a call to send and adhoc command to the queue?
    if args.adhoc:
        print("ADHOC is todo")
        return
    # sub to command topic if defined
    mqtt_broker.setAdhocCommands(config=mqttconfig, callback=mqtt_callback)
    log.debug(mqtt_broker)

    # get port
    portconfig = config["port"].copy()
    log.debug(f"portconfig: {portconfig}")
    port = get_port(portconfig)
    # port = get_port(porttype=porttype)
    if not port:
        log.error(f"No port for config '{portconfig}' found")
        raise ConfigError(f"No port for config '{portconfig}' found")

    # get protocol handler
    protocol = get_protocol(protocol=config["protocol"])

    # Get loop details
    loop = config.get("loop", "once")
    inDelay = False
    delayRemaining = loop
    doLoop = True

    try:
        # connect to port
        port.connect()
        while doLoop:
            # Start timer
            start_time = time()
            # loop through command list
            for command in config["commands"]:
                # process any adhoc commands first
                while len(ADHOC_COMMANDS) > 0:
                    log.debug(f"adhoc command list: {ADHOC_COMMANDS}")
                    adhoc_command = (
                        ADHOC_COMMANDS.popleft().decode()
                    )  # FIXME: decode to str #
                    log.info(f"Processing command: {adhoc_command}")
                    results = port.process_command(
                        command=adhoc_command, protocol=protocol
                    )
                    log.debug(f"results {results}")
                    # send to output processor(s)
                    # TODO sort outputs
                    output_results(
                        results=results,
                        command=config["mqttbroker"]["adhoc_commands"],
                        mqtt_broker=mqtt_broker,
                        fullconfig=config,
                    )
                # process 'normal' commands
                if not inDelay:
                    log.info(f"Processing command: {command}")
                    if "f_command" in command:
                        _command = command["f_command"]
                        _command = eval(_command)
                    else:
                        _command = command["command"]
                    results = port.process_command(command=_command, protocol=protocol)
                    log.debug(f"results {results}")
                    # send to output processor(s)
                    output_results(
                        results=results,
                        command=command,
                        mqtt_broker=mqtt_broker,
                        fullconfig=config,
                    )
                    # pause
                    # pause_time = config["command_pause"]
                    # log.debug(f"Sleeping for {pause_time}secs")
            if loop == "once":
                doLoop = False
            else:
                # Small pause to ....
                sleep(0.1)
                end_time = time()
                delayRemaining = delayRemaining - (end_time - start_time)
                if delayRemaining > 0 and not inDelay:
                    log.debug(
                        f"delaying for {loop}sec, delayRemaining: {delayRemaining}"
                    )
                    inDelay = True
                if delayRemaining < 0:
                    log.debug("setting inDelay to false")
                    inDelay = False
                    delayRemaining = loop
    except KeyboardInterrupt:
        print("KeyboardInterrupt")
    finally:
        # Disconnect port
        port.disconnect()
        # Disconnect mqtt
        mqtt_broker.stop()



import logging
import importlib
import pkgutil
import re

from ..helpers import key_wanted, get_kwargs

log = logging.getLogger("helpers")


def list_outputs():
    # print("outputs list outputs")
    pkgpath = __file__
    pkgpath = pkgpath[: pkgpath.rfind("/")]
    pkgpath += "/../outputs"
    # print(pkgpath)
    result = {}
    result["_command"] = "outputs help"
    result["_command_description"] = "List available output modules"
    for _, name, _ in pkgutil.iter_modules([pkgpath]):
        # print(name)
        try:
            _module_class = importlib.import_module("mppsolar.outputs." + name, ".")
            _module = getattr(_module_class, name)
            result[name] = (str(_module()), "", "")
        except ModuleNotFoundError as e:
            # log.error(f"Error in module {name}: {e}")
            result[f"{name}*"] = (f"ERROR: {e}", "", "")
        # print(_module())

    # print(result)
    return result


def get_output(output):
    """
    Take an output name
    attempt to find and instantiate the corresponding module
    """
    log.info(f"attempting to create output processor: {output}")
    try:
        output_module = importlib.import_module("mppsolar.outputs." + output, ".")
        output_class = getattr(output_module, output)
        return output_class()
    except ModuleNotFoundError as e:
        # perhaps raise a Powermon exception here??
        # maybe warn and keep going, only error if no outputs found?
        log.critical(f"No module found for output processor {output} Error: {e}")
    return None


def get_outputs(output_list):
    """
    Take a comma separated list of output names
    attempt to find and instantiate the corresponding module
    return array of modules
    """
    ops = []
    outputs = output_list.split(",")
    for output in outputs:
        op = get_output(output)
        if op is not None:
            ops.append(op)
    return ops


def output_results(results, command, mqtt_broker, fullconfig={}):
    # "normal command definition"
    # - command: QPIGS
    #   outputs:
    #   - name: screen
    #   - name: mqtt
    #     results_topic: results/qpigs
    #
    # or "adhoc_commands"
    #  topic: test/command_topic
    #  outputs:
    #  - name: screen

    # Check if an outputs section was supplied
    if "outputs" in command:
        outputs = command["outputs"]
    # if not default to screen
    else:
        outputs["name"] = "screen"
    for op in outputs:
        # filter = config.get("CONFIG", "filter")
        # log.debug(f"Using output filter: {filter}")
        if "name" in op:
            output = get_output(op["name"])
        else:
            output = get_output("screen")
        output.output(
            data=dict(results),
            config=op,
            mqtt_broker=mqtt_broker,
            fullconfig=fullconfig,
        )


def to_json(data, keep_case, excl_filter, filter):
    output = {}
    # Loop through responses
    for key in data:
        value = data[key]
        log.debug(f"value: {value}")
        if isinstance(value, list):
            value = data[key][0]
        # unit = data[key][1]
        # remove spaces
        key = key.replace(" ", "_")
        if not keep_case:
            # make lowercase
            key = key.lower()
        if key_wanted(key, filter, excl_filter):
            output[key] = value
    return output


def to_json_units(data, keep_case, excl_filter, filter):
    output = {}
    # Loop through responses
    for key in data:
        value = data[key]
        unit = None
        log.debug(f"value: {value}")
        if isinstance(value, list):
            value = data[key][0]
            unit = data[key][1]
        # remove spaces
        key = key.replace(" ", "_")
        if not keep_case:
            # make lowercase
            key = key.lower()
        if key_wanted(key, filter, excl_filter):
            if unit is None:
                output[key] = value
            else:
                # { "ac_output_voltage": { "value": 220, "unit": "V" }, ... }
                output[key] = {"value": value, "unit": unit}
    return output


def get_common_params(kwargs):
    data = get_kwargs(kwargs, "data")
    tag = get_kwargs(kwargs, "tag")
    keep_case = get_kwargs(kwargs, "keep_case")
    filter_ = get_kwargs(kwargs, "filter")
    if filter_ is not None:
        filter_ = re.compile(filter_)
    excl_filter = get_kwargs(kwargs, "excl_filter")
    if excl_filter is not None:
        excl_filter = re.compile(excl_filter)
    return data, tag, keep_case, filter_, excl_filter



import logging
import struct

from .abstractprotocol import AbstractProtocol
from .protocol_helpers import crc8


log = logging.getLogger("jkAbstractProtocol")

SOR = bytes.fromhex("55aaeb90")

COMMANDS = {
    "getInfo": {
        "name": "getInfo",
        "command_code": "97",
        "record_type": "3",
        "description": "BLE Device Information inquiry",
        "help": " -- queries the ble device information",
        "type": "QUERY",
        "response_type": "POSITIONAL",
        "response": [
            ["Hex2Str", 4, "Header", ""],
            ["Hex2Str", 1, "Record Type", ""],
            ["Hex2Int", 1, "Record Counter", ""],
            ["Hex2Ascii", 16, "Device Model", ""],
            ["Hex2Ascii", 8, "Hardware Version", ""],
            ["Hex2Ascii", 8, "Software Version", ""],
            ["uptime", 4, "Up Time", ""],
            ["Hex2Int", 4, "Power-on Times", ""],
            ["Hex2Ascii", 16, "Device Name", ""],
            ["Hex2Ascii", 16, "Device Passcode", ""],
            ["Hex2Ascii", 8, "Manufacturing Date", ""],
            ["Hex2Ascii", 11, "Serial Number", ""],
            ["Hex2Ascii", 5, "Passcode", ""],
            ["Hex2Ascii", 16, "User Data", ""],
            ["Hex2Ascii", 16, "Setup Passcode", ""],
            ["discard", 672, "unknown", ""],
        ],
        "test_responses": [
            bytes.fromhex(
                "55aaeb9003f14a4b2d42324132345300000000000000332e300000000000332e322e330000000876450004000000506f7765722057616c6c203100000000313233340000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000c2"
            ),
            bytes.fromhex(
                "55aaeb9003b54a4b2d42443641323053313050000000342e300000000000342e312e37000000541d1600040000004e6f7468696e67204a4b31000000000031323334000000000000000000000000323030373038000032303036323834303735000000000000496e707574205573657264617461000031323334353600000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000c4"
            ),
        ],
    },
}


class jkAbstractProtocol(AbstractProtocol):
    """
    JKAbstractProtocol - Abstract Handler for JKBMS communication
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__()
        self._protocol_id = b"JK"
        self.COMMANDS = COMMANDS
        self.STATUS_COMMANDS = [
            "",
        ]
        self.SETTINGS_COMMANDS = [
            "getInfo",
        ]
        self.DEFAULT_COMMAND = "getInfo"

    def get_full_command(self, command) -> bytes:
        """
        Override the default get_full_command as its different for JK
        """
        # getInfo = b'\xaa\x55\x90\xeb\x97\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x11'
        log.info(
            f"Using protocol {self._protocol_id} with {len(self.COMMANDS)} commands"
        )
        # These need to be set to allow other functions to work`
        self._command = command
        self._command_defn = self.get_command_defn(command)
        # log.debug(f"self._command = {self._command}, self._command_defn = {self._command_defn}")
        log.debug(f"self._command = {self._command}")
        # End of required variables setting
        if self._command_defn is None:
            # Maybe return a default here?
            log.debug("No command_defn found")
            return None
        if "command_code" in self._command_defn:
            # full command is 20 bytes long
            cmd = bytearray(20)
            # starts with \xaa\x55\x90\xeb
            cmd[0:4] = bytes.fromhex("aa5590eb")
            log.debug(f"cmd with SOR: {cmd}")
            # then has command code
            cmd[4] = int(self._command_defn["command_code"], 16)
            if self._command_defn["type"] == "SETTER":
                cmd[5] = 0x04
                value = struct.pack("<h", int(float(self._command_value) * 1000))
                cmd[6] = value[0]
                cmd[7] = value[1]
            log.debug(f"cmd with command code: {cmd}")
            cmd[-1] = crc8(cmd)
            log.debug(f"cmd with crc: {cmd}")
            return cmd
        return None

    def get_command_defn(self, command):
        log.debug(f"get_command_defn for: {command}")
        if command is None:
            log.debug("command is None")
            return None
        return super().get_command_defn(command)

    def get_responses(self, response):
        """
        Override the default get_responses as its different for JK
        """
        responses = []
        if (
            self._command_defn is not None
            and self._command_defn["response_type"] == "POSITIONAL"
        ):
            # Have a POSITIONAL type response, so need to break it up...
            # example defn :
            # "response": [
            #   ["discard", 1, "start flag", ""],
            #   ["discard", 1, "module address", ""],
            #   ["discard", 1, "command id", ""],
            #   ["discard", 1, "data length", ""],
            # ]
            # example response data b"\xa5\x01\x90\x08\x02\x10\x00\x00uo\x03\xbc\xf3",
            for defn in self._command_defn["response"]:
                log.debug(f"Got defn: {defn}")
                if defn[0].startswith("lookup"):
                    responses.append("lookup")
                    continue
                size = defn[1]
                item = response[:size]
                responses.append(item)
                response = response[size:]
            if len(response) > 0:
                responses.append(response)
            return responses
        else:
            return bytearray(response)

    def is_record_start(self, record):
        if record.startswith(SOR):
            log.debug("SOR found in record")
            return True
        return False

    def wipe_to_start(self, record):
        sor_loc = record.find(SOR)
        if sor_loc == -1:
            log.debug("SOR not found in record")
            return bytearray()
        return record[sor_loc:]

    def is_record_correct_type(self, record, type):
        if len(record) < len(SOR):
            return False
        if record[len(SOR)] == int(type):
            log.debug(f"Record is type {type}")
            return True
        return False

    def is_record_complete(self, record):
        """"""
        # check record starts with 'SOR'
        if not self.is_record_start(record):
            log.debug("No SOR found in record looking for completeness")
            return False
        # check that length one of the valid lengths (300, 320)
        if len(record) == 300 or len(record) == 320:
            # check the crc/checksum is correct for the record data
            crc = record[299]
            calcCrc = crc8(record[:-1])
            # print (crc, calcCrc)
            if crc == calcCrc:
                log.debug("Record CRC is valid")
                return True
        return False



# shamelessly stolen from ccrisan https://github.com/qtoggle/qtoggleserver-mppsolar/blob/master/qtoggleserver/mppsolar/io.py
import logging
import random
import re

from .baseio import BaseIO
from ..helpers import get_kwargs

log = logging.getLogger("TestIO")


class TestIO(BaseIO):
    def __init__(self, *args, **kwargs) -> None:
        self._test_data = b"(230.0 50.0 0030 42.0 54.0 56.4 46.0 60 0 0 2 0 0 0 0 0 1 1 0 0 1 0 54.0 0 1 000\x9E\x60\r"
        self._counter = 0
        self._device = get_kwargs(kwargs, "device_path")

    def send_and_receive(self, *args, **kwargs) -> dict:
        # full_command = get_kwargs(kwargs, "full_command")
        command_defn = get_kwargs(kwargs, "command_defn")

        if command_defn is not None:
            # Have test data defined, so use that
            number_of_test_responses = len(command_defn["test_responses"])
            # have we specified which test response to use?
            check = re.match("test(\d+)", self._device, flags=re.IGNORECASE)  # noqa: W605
            if check:
                # use the specified response if that one exists
                try:
                    desired_test_response = int(check.group(1))
                except ValueError:
                    desired_test_response = 0
                if desired_test_response < number_of_test_responses:
                    self._test_data = command_defn["test_responses"][desired_test_response]
                else:
                    log.warn(
                        f"Test response requested: {desired_test_response} exceeds highest id available (IDs start at 0): {number_of_test_responses-1}, returning random one"
                    )
                    self._test_data = command_defn["test_responses"][
                        random.randrange(number_of_test_responses)
                    ]
            else:
                # just use a random test response
                self._test_data = command_defn["test_responses"][
                    random.randrange(number_of_test_responses)
                ]
        else:
            # No test responses defined
            log.warn("Testing a command with no test responses defined")
        response = self._test_data
        log.debug(f"Raw response {response}")
        return response



import logging
from datetime import datetime

import psycopg2 as psycopg2
from psycopg2._json import Json
from psycopg2.extensions import register_adapter

from . import to_json, get_common_params
from .baseoutput import baseoutput
from ..helpers import get_kwargs

log = logging.getLogger("postgres")

"""
To use, run `sudo apt-get install libpq-dev`

Creating table in Postgresql in advance:

-- auto-generated definition
create table mppsolar
(
    id      serial
        constraint mppsolar_pk
            primary key,
    command varchar,
    data    json,
    updated timestamp
);

create index mppsolar_command_updated_index
    on mppsolar (command, updated);


"""


class postgres(baseoutput):
    def __str__(self):
        return "outputs all the results to PostgresSQL"

    def __init__(self, *args, **kwargs) -> None:
        log.debug(f"__init__: kwargs {kwargs}")
        register_adapter(dict, Json)

    def output(self, *args, **kwargs):
        (data, tag, keep_case, filter_, excl_filter) = get_common_params(kwargs)

        postgres_url = get_kwargs(kwargs, "postgres_url")
        log.debug(f"Connecting to {postgres_url}")
        conn = psycopg2.connect(postgres_url)

        msgs = []
        # Remove command and _command_description
        # cmd = data.pop("_command", None)
        data.pop("_command_description", None)
        data.pop("raw_response", None)
        # if tag is None:
        #     tag = cmd
        output = to_json(data, keep_case, excl_filter, filter_)

        log.debug(output)
        msgs.append(output)
        inserted = 0
        now = datetime.now().astimezone().replace(microsecond=0).isoformat()
        try:
            for msg in msgs:
                command = msg.pop("_command")
                msg['updated'] = now
                log.debug(conn)
                cursor = conn.cursor()
                log.debug(cursor)
                cursor.execute('insert into mppsolar (command,data, updated) values (%s,%s,%s)', (command, msg, now))
                conn.commit()
                inserted += 1
                cursor.close()
            log.debug(f"inserted {inserted} docs")
        except Exception as e:
            log.error(f"Postgres error {e}")
        finally:
            conn.close()
        return msgs



import datetime
import logging
import re

import pymongo as pymongo

from . import to_json
from .baseoutput import baseoutput
from ..helpers import get_kwargs
# from ..helpers import key_wanted

log = logging.getLogger("mongo")


class mongo(baseoutput):
    def __str__(self):
        return "outputs all the results to MongoDB"

    def __init__(self, *args, **kwargs) -> None:
        log.debug(f"__init__: kwargs {kwargs}")

    def output(self, *args, **kwargs):
        data = get_kwargs(kwargs, "data")
        # tag = get_kwargs(kwargs, "tag")
        keep_case = get_kwargs(kwargs, "keep_case")
        filter = get_kwargs(kwargs, "filter")
        if filter is not None:
            filter = re.compile(filter)
        excl_filter = get_kwargs(kwargs, "excl_filter")
        if excl_filter is not None:
            excl_filter = re.compile(excl_filter)

        mongo_url = get_kwargs(kwargs, "mongo_url")
        mongo_database = get_kwargs(kwargs, "mongo_db", "mppsolar")
        log.debug(f"Connecting to {mongo_url} / {mongo_database}")
        client = pymongo.MongoClient(mongo_url)
        db = client[mongo_database]

        msgs = []
        # Remove command and _command_description
        # cmd = data.pop("_command", None)
        data.pop("_command_description", None)
        data.pop("raw_response", None)
        # if tag is None:
        #     tag = cmd
        output = to_json(data, keep_case, excl_filter, filter)

        log.debug(output)
        msgs.append(output)
        inserted = 0
        try:
            for msg in msgs:
                col = msg.pop("_command")
                msg['updated'] = datetime.datetime.now()
                result = db[col].insert_one(msg)
                if result is not None:
                    log.debug(result.inserted_id)
                    inserted += 1
            log.debug(f"inserted {inserted} docs")
        except pymongo.errors.ServerSelectionTimeoutError as dbe:
            log.error(f"Mongo error {dbe}")
        return msgs



#!/usr/bin/env python3
import logging
import importlib

log = logging.getLogger("helpers")


def get_kwargs(kwargs, key, default=None):
    if key not in kwargs or not kwargs[key]:
        return default
    return kwargs[key]


def key_wanted(key, filter=None, excl_filter=None):
    # remove any specifically excluded keys
    if excl_filter is not None and excl_filter.search(key):
        # log.debug(f"key_wanted: key {key} matches excl_filter {excl_filter} so key excluded")
        return False
    if filter is None:
        # log.debug(
        #    f"key_wanted: No filter and key {key} not excluded by excl_filter {excl_filter} so key wanted"
        # )
        return True
    elif filter.search(key):
        # log.debug(
        #    f"key_wanted: key {key} matches filter {filter} and not excl_filter {excl_filter} so key wanted"
        # )
        return True
    else:
        # log.debug(f"key_wanted: key {key} does not match filter {filter} so key excluded")
        return False


def get_value(_list, _index):
    """
    get the value from _list or return None if _index is out of bounds
    """
    # print(_list, len(_list))
    if _index >= len(_list):
        return None
    return _list[_index]


def get_resp_defn(key, defns):
    """
    look for a definition for the supplied key
    """
    # print(key, defns)
    if not key:
        return None
    if type(key) is bytes:
        try:
            key = key.decode("utf-8")
        except UnicodeDecodeError:
            log.info(f"key decode error for {key}")
    for defn in defns:
        if key == defn[0]:
            # print(key, defn)
            return defn
    # did not find definition for this key
    log.info(f"No defn found for {key} key")
    return [key, key, "", ""]


def get_device_class(device_type=None):
    """
    Take a device type string
    attempt to find and instantiate the corresponding module
    return class if found, otherwise return None
    """
    if device_type is None:
        return None
    device_type = device_type.lower()
    try:
        device_module = importlib.import_module("mppsolar.devices." + device_type, ".")
    except ModuleNotFoundError as e:
        # perhaps raise a mppsolar exception here??
        log.critical(f"Error loading device {device_type}: {e}")
        return None
    device_class = getattr(device_module, device_type)
    return device_class


def getMaxLen(d):
    _maxLen = 0
    for i in d:
        if type(i) == list:
            i = i[0]
        if len(i) > _maxLen:
            _maxLen = len(i)
    return _maxLen


def pad(text, length):
    if len(text) > length:
        return text
    return text.ljust(length, " ")



import logging

from .jkabstractprotocol import jkAbstractProtocol


log = logging.getLogger("jkv11")

NEW_COMMANDS = {
    "getCellData": {
        "name": "getCellData",
        "command_code": "96",
        "record_type": "2",
        "description": "BLE Cell Data inquiry",
        "help": " -- queries the ble device for the cell data",
        "type": "QUERY",
        "response_type": "POSITIONAL",
        "response": [
            ["Hex2Str", 4, "Header", ""],
            ["Hex2Str", 1, "Record_Type", ""],
            ["Hex2Int", 1, "Record_Counter", ""],
            ["LittleHex2Short:r/1000", 2, "Voltage_Cell01", "V"],
            ["LittleHex2Short:r/1000", 2, "Voltage_Cell02", "V"],
            ["LittleHex2Short:r/1000", 2, "Voltage_Cell03", "V"],
            ["LittleHex2Short:r/1000", 2, "Voltage_Cell04", "V"],
            ["LittleHex2Short:r/1000", 2, "Voltage_Cell05", "V"],
            ["LittleHex2Short:r/1000", 2, "Voltage_Cell06", "V"],
            ["LittleHex2Short:r/1000", 2, "Voltage_Cell07", "V"],
            ["LittleHex2Short:r/1000", 2, "Voltage_Cell08", "V"],
            ["LittleHex2Short:r/1000", 2, "Voltage_Cell09", "V"],
            ["LittleHex2Short:r/1000", 2, "Voltage_Cell10", "V"],
            ["LittleHex2Short:r/1000", 2, "Voltage_Cell11", "V"],
            ["LittleHex2Short:r/1000", 2, "Voltage_Cell12", "V"],
            ["LittleHex2Short:r/1000", 2, "Voltage_Cell13", "V"],
            ["LittleHex2Short:r/1000", 2, "Voltage_Cell14", "V"],
            ["LittleHex2Short:r/1000", 2, "Voltage_Cell15", "V"],
            ["LittleHex2Short:r/1000", 2, "Voltage_Cell16", "V"],
            ["LittleHex2Short:r/1000", 2, "Voltage_Cell17", "V"],
            ["LittleHex2Short:r/1000", 2, "Voltage_Cell18", "V"],
            ["LittleHex2Short:r/1000", 2, "Voltage_Cell19", "V"],
            ["LittleHex2Short:r/1000", 2, "Voltage_Cell20", "V"],
            ["LittleHex2Short:r/1000", 2, "Voltage_Cell21", "V"],
            ["LittleHex2Short:r/1000", 2, "Voltage_Cell22", "V"],
            ["LittleHex2Short:r/1000", 2, "Voltage_Cell23", "V"],
            ["LittleHex2Short:r/1000", 2, "Voltage_Cell24", "V"],
            ["LittleHex2Short:r/1000", 2, "Voltage_Cell25", "V"],
            ["LittleHex2Short:r/1000", 2, "Voltage_Cell26", "V"],
            ["LittleHex2Short:r/1000", 2, "Voltage_Cell27", "V"],
            ["LittleHex2Short:r/1000", 2, "Voltage_Cell28", "V"],
            ["LittleHex2Short:r/1000", 2, "Voltage_Cell29", "V"],
            ["LittleHex2Short:r/1000", 2, "Voltage_Cell30", "V"],
            ["LittleHex2Short:r/1000", 2, "Voltage_Cell31", "V"],
            ["LittleHex2Short:r/1000", 2, "Voltage_Cell32", "V"],
            ["discard", 4, "discard1", ""],
            ["LittleHex2Short:r/1000", 2, "Average_Cell_Voltage", "V"],
            ["LittleHex2Short:r/1000", 2, "Delta_Cell_Voltage", "V"],
            ["LittleHex2Short:r/1000", 2, "Current_Balancer", ""],
            ["LittleHex2Short:r/1000", 2, "Resistance_Cell01", "Ohm"],
            ["LittleHex2Short:r/1000", 2, "Resistance_Cell02", "Ohm"],
            ["LittleHex2Short:r/1000", 2, "Resistance_Cell03", "Ohm"],
            ["LittleHex2Short:r/1000", 2, "Resistance_Cell04", "Ohm"],
            ["LittleHex2Short:r/1000", 2, "Resistance_Cell05", "Ohm"],
            ["LittleHex2Short:r/1000", 2, "Resistance_Cell06", "Ohm"],
            ["LittleHex2Short:r/1000", 2, "Resistance_Cell07", "Ohm"],
            ["LittleHex2Short:r/1000", 2, "Resistance_Cell08", "Ohm"],
            ["LittleHex2Short:r/1000", 2, "Resistance_Cell09", "Ohm"],
            ["LittleHex2Short:r/1000", 2, "Resistance_Cell10", "Ohm"],
            ["LittleHex2Short:r/1000", 2, "Resistance_Cell11", "Ohm"],
            ["LittleHex2Short:r/1000", 2, "Resistance_Cell12", "Ohm"],
            ["LittleHex2Short:r/1000", 2, "Resistance_Cell13", "Ohm"],
            ["LittleHex2Short:r/1000", 2, "Resistance_Cell14", "Ohm"],
            ["LittleHex2Short:r/1000", 2, "Resistance_Cell15", "Ohm"],
            ["LittleHex2Short:r/1000", 2, "Resistance_Cell16", "Ohm"],
            ["LittleHex2Short:r/1000", 2, "Resistance_Cell17", "Ohm"],
            ["LittleHex2Short:r/1000", 2, "Resistance_Cell18", "Ohm"],
            ["LittleHex2Short:r/1000", 2, "Resistance_Cell19", "Ohm"],
            ["LittleHex2Short:r/1000", 2, "Resistance_Cell20", "Ohm"],
            ["LittleHex2Short:r/1000", 2, "Resistance_Cell21", "Ohm"],
            ["LittleHex2Short:r/1000", 2, "Resistance_Cell22", "Ohm"],
            ["LittleHex2Short:r/1000", 2, "Resistance_Cell23", "Ohm"],
            ["LittleHex2Short:r/1000", 2, "Resistance_Cell24", "Ohm"],
            ["LittleHex2Short:r/1000", 2, "Resistance_Cell25", "Ohm"],
            ["LittleHex2Short:r/1000", 2, "Resistance_Cell26", "Ohm"],
            ["LittleHex2Short:r/1000", 2, "Resistance_Cell27", "Ohm"],
            ["LittleHex2Short:r/1000", 2, "Resistance_Cell28", "Ohm"],
            ["LittleHex2Short:r/1000", 2, "Resistance_Cell29", "Ohm"],
            ["LittleHex2Short:r/1000", 2, "Resistance_Cell30", "Ohm"],
            ["LittleHex2Short:r/1000", 2, "Resistance_Cell31", "Ohm"],
            ["LittleHex2Short:r/1000", 2, "Resistance_Cell32", "Ohm"],
            ["discard", 6, "discard2", ""],
            ["LittleHex2UInt:r/1000", 4, "Battery_Voltage", "V"],
            ["LittleHex2UInt", 4, "Battery_Power", "W"],
            ["LittleHex2Int:r/1000", 4, "Balance_Current", "A"],  # signed int32
            # ["discard", 8, "discard3", ""],
            ["LittleHex2Short:r/10", 2, "Battery_T1", "°C"],
            ["LittleHex2Short:r/10", 2, "Battery_T2", "°C"],
            ["LittleHex2Short:r/10", 2, "MOS_Temp", "°C"],
            ["discard", 4, "discard4", ""],  # discard4
            ["discard", 1, "discard4_1", ""],  # added
            ["Hex2Int", 1, "Percent_Remain", "%"],
            ["LittleHex2UInt:r/1000", 4, "Capacity_Remain", "Ah"],  # Unknown6+7
            ["LittleHex2UInt:r/1000", 4, "Nominal_Capacity", "Ah"],  # Unknown8+9
            ["LittleHex2UInt", 4, "Cycle_Count", ""],
            # ["discard", 2, "Unknown10", ""],
            # ["discard", 2, "Unknown11", ""],
            ["LittleHex2UInt:r/1000", 4, "Cycle_Capacity", "Ah"],  # Unknown10+11
            ["discard", 2, "Unknown12", ""],
            ["discard", 2, "Unknown13", ""],
            ["uptime", 3, "Time", ""],
            ["discard", 2, "Unknown15", ""],
            ["discard", 2, "Unknown16", ""],
            ["discard", 2, "Unknown17", ""],
            ["discard", 12, "discard6", ""],
            ["discard", 2, "Unknown18", ""],
            ["discard", 2, "Unknown19", ""],
            ["discard", 2, "Unknown20", ""],
            ["LittleHex2Short:r/1000", 2, "Current_Charge", ""],  # Unknown21
            ["LittleHex2Short:r/1000", 2, "Current_Discharge", ""],  # Unknown22
            ["discard", 2, "Unknown23", ""],
            ["discard", 2, "Unknown24", ""],
            ["discard", 2, "Unknown25", ""],
            ["discard", 2, "Unknown26", ""],
            ["discard", 2, "Unknown27", ""],
            ["discard", 2, "Unknown28", ""],
            ["discard", 2, "Unknown29", ""],
            ["discard", 93, "Unknown30", ""],
        ],
        "test_responses": [
        ],
    },
    "setCellOVP": {
        "name": "setCellOVP",
        "command_code": "04",
        "record_type": "2",
        "description": "Set cell over voltage protection",
        "help": " -- example setCellOVP3.65",
        "type": "SETTER",
        "response_type": "POSITIONAL",
        "response": [["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}]],
        "test_responses": [
            bytes.fromhex(
                "55aaeb9002b52e0d280dfa0c2e0d2f0d220d220d130d190d1d0d1d0d170d1f0d160dfb0c1f0d00000000000000000000000000000000ffff00001c0d350004029b00c600a000b300bc00cc00be00b100b4002d013d01b000a100ab00b200ad0000000000000000000000000000000000000000000000bcd1000000000000000000001e0116013c010000000000636b0c0300400d030000000000dc4d010064000000781e16000101480a000000000000000000000000070101000000980400000000260141400000000037feffff00000001000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000080"
            ),
        ],
        "regex": "setCellOVP(\\d\\.\\d*)$",
    },
}


class jkv11(jkAbstractProtocol):
    """
    JKv11 - Handler for JKBMS 2 byte data communication
         - e.g. ASAS = ??V
    """
    def __str__(self):
        return "JKBMS BLE communication protocol handler software v11.x"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__()
        self._protocol_id = b"JKv11"
        self.COMMANDS.update(NEW_COMMANDS)
        self.STATUS_COMMANDS = [
            "getCellData",
        ]
        self.SETTINGS_COMMANDS = [
            "getInfo",
        ]
        self.DEFAULT_COMMAND = "getCellData"



import logging
import random

from ..helpers import get_kwargs
from .port import Port

# import re


log = logging.getLogger("Test")


class testport(Port):
    def __init__(self, *args, **kwargs) -> None:
        # self._test_data = None
        # self._counter = 0
        # self._device = get_kwargs(kwargs, "device_path")
        log.debug(f"Initializing test port args:{args}, kwargs: {kwargs}")

    def connect(self) -> None:
        log.debug("Test port connected")
        return

    def send_and_receive(self, *args, **kwargs) -> dict:
        # full_command = get_kwargs(kwargs, "full_command")
        command_defn = get_kwargs(kwargs, "command_defn")

        if command_defn is not None:
            # Have test data defined, so use that
            number_of_test_responses = len(command_defn["test_responses"])
            self._test_data = command_defn["test_responses"][
                random.randrange(number_of_test_responses)
            ]
        else:
            # No test responses defined
            log.warn("Testing a command with no test responses defined")
            self._test_data = None
        response = self._test_data
        log.debug(f"Raw response {response}")
        return response



# import importlib
import logging
from abc import ABC

from mppsolar.helpers import get_kwargs
from mppsolar.inout import get_port
from mppsolar.protocols import get_protocol

PORT_TYPE_UNKNOWN = 0
PORT_TYPE_TEST = 1
PORT_TYPE_USB = 2
PORT_TYPE_ESP32 = 4
PORT_TYPE_SERIAL = 8
PORT_TYPE_JKBLE = 16
PORT_TYPE_MQTT = 32
PORT_TYPE_VSERIAL = 64
PORT_TYPE_DALYSERIAL = 128

# log = None
log = logging.getLogger("device")


class AbstractDevice(ABC):
    """
    Abstract device class
    """

    def __init__(self, *args, **kwargs):
        global log
        # self._protocol = None
        # self._protocol_class = None
        # self._port = None
        log.debug(f"__init__ args {args}")
        log.debug(f"__init__ kwargs {kwargs}")
        self._name = get_kwargs(kwargs, "name")
        self._port = get_port(**kwargs)
        self._protocol = get_protocol(get_kwargs(kwargs, "protocol"))
        log.debug(
            f"__init__ name {self._name}, port {self._port}, protocol {self._protocol}"
        )

    def __str__(self):
        """
        Build a printable representation of this class
        """
        return f"{self._classname} device - name: {self._name}, port: {self._port}, protocol: {self._protocol}"

    def run_command(self, command) -> dict:
        """
        generic method for running a 'raw' command
        """
        log.info(f"Running command {command}")

        if self._protocol is None:
            log.error("Attempted to run command with no protocol defined")
            return {"ERROR": ["Attempted to run command with no protocol defined", ""]}
        if self._port is None:
            log.error(
                f"No communications port defined - unable to run command {command}"
            )
            return {
                "ERROR": [
                    f"No communications port defined - unable to run command {command}",
                    "",
                ]
            }

        if command == "list_commands":
            return self._protocol.list_commands()
        if command == "get_status":
            return self.get_status()
        if command == "get_settings":
            return self.get_settings()
        if not command:
            command = self._protocol.DEFAULT_COMMAND

        # Send command and receive data
        full_command = self._protocol.get_full_command(command)
        log.info(f"full command {full_command} for command {command}")
        if full_command is None:
            log.error(
                f"full_command not found for {command} in protocol {self._protocol._protocol_id}"
            )
            return {
                "ERROR": [
                    f"full_command not found for {command} in protocol {self._protocol._protocol_id}",
                    "",
                ]
            }

        # Band-aid solution, need to reduce what is sent
        raw_response = self._port.send_and_receive(
            command=command,
            full_command=full_command,
            protocol=self._protocol,
            command_defn=self._protocol.get_command_defn(command),
        )
        log.debug(f"Send and Receive Response {raw_response}")

        # Handle errors
        # Maybe there should a decode for ERRORs and WARNINGS...
        # Some inverters return the command if the command is unknown:
        if raw_response == full_command:
            return {
                "ERROR": [
                    f"Inverter returned the command string for {command} - the inverter didnt recognise this command",
                    "",
                ]
            }
        # dict is returned on exception
        # if isinstance(raw_response, dict):
        #     return raw_response

        # Decode response
        decoded_response = self._protocol.decode(raw_response, command)
        log.info(f"Decoded response {decoded_response}")

        return decoded_response

    def get_status(self) -> dict:
        # Run all the commands that are defined as status from the protocol definition
        data = {}
        for command in self._protocol.STATUS_COMMANDS:
            data.update(self.run_command(command))
        return data

    def get_settings(self) -> dict:
        # Run all the commands that are defined as settings from the protocol definition
        data = {}
        for command in self._protocol.SETTINGS_COMMANDS:
            data.update(self.run_command(command))
        return data



# shamelessly stolen from ccrisan https://github.com/qtoggle/qtoggleserver-mppsolar/blob/master/qtoggleserver/mppsolar/io.py
import logging
import os
import time

from .baseio import BaseIO
from ..helpers import get_kwargs

log = logging.getLogger("HIDFullIO")


class HIDFullIO(BaseIO):
    def __init__(self, device_path: str) -> None:
        # self._fd = os.open(device_path, flags=os.O_RDWR | os.O_NONBLOCK)
        self._device = device_path

    def send_and_receive(self, *args, **kwargs) -> dict:
        full_command = get_kwargs(kwargs, "full_command")
        response_line = bytes()
        usb0 = None
        try:
            usb0 = os.open(self._device, os.O_RDWR | os.O_NONBLOCK)
        except Exception as e:
            log.debug("USB open error: {}".format(e))
            return {"ERROR": ["USB open error: {}".format(e), ""]}
        # Send the command to the open usb connection
        to_send = full_command
        try:
            log.debug(f"length of to_send: {len(to_send)}")
        except:  # noqa: E722
            import pdb

            pdb.set_trace()

        # Send all at once
        log.debug("1 chunk send")
        time.sleep(0.35)
        os.write(usb0, to_send)

        time.sleep(0.25)
        # Read from the usb connection
        # try to a max of 100 times
        for x in range(100):
            # attempt to deal with resource busy and other failures to read
            try:
                time.sleep(0.15)
                r = os.read(usb0, 256)
                response_line += r
            except Exception as e:
                log.debug("USB read error: {}".format(e))
            # Finished is \r is in byte_response
            if bytes([13]) in response_line:
                # remove anything after the \r
                response_line = response_line[: response_line.find(bytes([13])) + 1]
                break
        log.debug("usb response was: %s", response_line)
        os.close(usb0)
        return response_line



import logging
import re

from .mqtt import mqtt
from ..helpers import get_kwargs
from ..helpers import key_wanted

log = logging.getLogger("tag_mqtt")


class tag_mqtt(mqtt):
    def __str__(self):
        return "outputs the to the supplied mqtt broker using the supplied tag as the topic: eg {tag}/max_charger_range 120.0"

    def __init__(self, *args, **kwargs) -> None:
        log.debug(f"__init__: kwargs {kwargs}")

    def build_msgs(self, *args, **kwargs):
        data = get_kwargs(kwargs, "data")
        tag = get_kwargs(kwargs, "tag")
        keep_case = get_kwargs(kwargs, "keep_case")
        _topic = get_kwargs(kwargs, "topic", default="mpp-solar")
        if tag is None:
            tag = _topic
        filter = get_kwargs(kwargs, "filter")
        if filter is not None:
            filter = re.compile(filter)
        excl_filter = get_kwargs(kwargs, "excl_filter")
        if excl_filter is not None:
            excl_filter = re.compile(excl_filter)

        # Build array of Influx Line Protocol II messages
        # Message format is: mpp-solar,command=QPGS0 max_charger_range=120.0
        #                    mpp-solar,command=inverter2 parallel_instance_number="valid"
        #                    measurement,tag_set field_set
        msgs = []
        # Remove command and _command_description
        cmd = data.pop("_command", None)
        data.pop("_command_description", None)
        data.pop("raw_response", None)
        if tag is None:
            tag = cmd
        # Loop through responses
        for key in data:
            value = data[key][0]
            # remove spaces
            key = key.replace(" ", "_")
            if not keep_case:
                # make lowercase
                key = key.lower()
            if key_wanted(key, filter, excl_filter):
                msg = {
                    "topic": f"{tag}/{key}",
                    "payload": value,
                }
                msgs.append(msg)
        return msgs



import logging
import serial
import time

from .port import Port
from ..helpers import get_kwargs

log = logging.getLogger("usbserial")


class serialport(Port):
    def __init__(self, *args, **kwargs) -> None:
        log.debug(f"Initializing usbserial port args:{args}, kwargs: {kwargs}")
        self._config = get_kwargs(kwargs, "config")
        self.port = None
        self.error = None

    def connect(self) -> None:
        log.debug("usbserial port connecting")
        port = self._config.get("path")
        baud = self._config.get("baud")
        try:
            self.port = serial.Serial(port, baud, timeout=1, write_timeout=1)
        except Exception as e:
            log.warning(f"Error openning serial port: {e}")
            self.error = e
        return

    def disconnect(self) -> None:
        log.debug("usbserial port disconnecting")
        if self.port is not None:
            self.port.close()
        return

    def send_and_receive(self, *args, **kwargs) -> dict:
        full_command = get_kwargs(kwargs, "full_command")
        response_line = None
        log.debug(f"port {self.port}")
        if self.port is None:
            log.error("Port not available")
            return {"ERROR": [f"Serial command execution failed {self.error}", ""]}
        try:
            log.debug("Executing command via usbserial...")
            self.port.reset_input_buffer()
            self.port.reset_output_buffer()
            c = self.port.write(full_command)
            log.debug(f"Wrote {c} bytes")
            self.port.flush()
            time.sleep(0.1)  # give serial port time to receive the data
            response_line = self.port.read_until(b"\r")
            log.debug("serial response was: %s", response_line)
            return response_line
        except Exception as e:
            log.warning(f"Serial read error: {e}")
        log.info("Command execution failed")
        return {"ERROR": ["Serial command execution failed", ""]}
